#include "../support/logger.h"
#include "generator.h"

#include <string.h>

#define DEFAULT_ERROR 5
#define Print(...) fprintf(file, ##__VA_ARGS__)

static FILE * file;

void PrintDerivative(Derivative * derivative) {
    if (!(derivative->value->type == ID_EXPRESSION && !strcmp(derivative->value->id , "x"))) {
        Print("function_handle");
    }
    Print("(diff(%s, ", derivative->function);
    Print("x, %d)", *derivative->grade);
    Print(")");
    if (!(derivative->value->type == ID_EXPRESSION && !strcmp(derivative->value->id , "x"))) {
        Print("(");
        PrintExpression(derivative->value);
        Print(")");
    }
}

void PrintIntegrate(Integrate * integrate) {
    Print("quadv(%s, ", integrate->function);
    PrintExpression(integrate->lvalue);
    Print(", ");
    PrintExpression(integrate->rvalue);
    Print(", %f)", integrate->error == NULL ? DEFAULT_ERROR : *integrate->error);
}

void PrintEvaluate(Evaluate * evaluate) {
    Print("%s", evaluate->function); 
    Print("(");
    PrintExpression(evaluate->value);
    Print(")");
}

void PrintExpression(Expression * expression) {
    switch (expression->type) {
        case DERIVATIVE_EXPRESSION:
            PrintDerivative(expression->derivative);
            break;
        case INTEGRATE_EXPRESSION:
            PrintIntegrate(expression->integrate);
            break;
        case EVALUATE_EXPRESSION:
            PrintEvaluate(expression->evaluate);
            break;
        case NEGATIVE_EXPRESSION:
            Print(" - ");
            PrintExpression(expression->leftExpression);
            break;
        case PARENTHESIS_EXPRESSION:
            Print("(");
            PrintExpression(expression->leftExpression);
            Print(")");
            break;
        case NUMBER_EXPRESSION:
            Print("%f", expression->value);
            break;
        case ADD_EXPRESSION:
            PrintExpression(expression->leftExpression);
            Print(" + ");
            PrintExpression(expression->rightExpression);
            break;
        case SUB_EXPRESSION:
            PrintExpression(expression->leftExpression);
            Print(" - ");
            PrintExpression(expression->rightExpression);
            break;
        case MUL_EXPRESSION:
            PrintExpression(expression->leftExpression);
            Print(" * ");
            PrintExpression(expression->rightExpression);
            break;
        case DIV_EXPRESSION:
            PrintExpression(expression->leftExpression);
            Print(" / ");
            PrintExpression(expression->rightExpression);
            break;
        case POW_EXPRESSION:
            PrintExpression(expression->leftExpression);
            Print(".^");
            PrintExpression(expression->rightExpression);
            break;
        case ID_EXPRESSION:
            Print("%s", expression->id);
            break;
        default:
            abort();
    }
}

void PrintDomain(Domain * domain) {
    switch (domain->type) {
    case GREATER_DOMAIN:
        PrintExpression(domain->left);
        Print(" > ");
        PrintExpression(domain->right);
        break;
    case GREATER_EQUAL_DOMAIN:
        PrintExpression(domain->left);
        Print(" >= ");
        PrintExpression(domain->right);
        break;
    case LESSER_EQUAL_DOMAIN:
        PrintExpression(domain->left);
        Print(" <= ");
        PrintExpression(domain->right);
        break;
    case LESSER_DOMAIN:
        PrintExpression(domain->left);
        Print(" < ");
        PrintExpression(domain->right);
        break;
    case EQUAL_DOMAIN:
        PrintExpression(domain->left);
        Print(" == ");
        PrintExpression(domain->right);
        break;
    case NOT_EQUAL_DOMAIN:
        PrintExpression(domain->left);
        Print(" != ");
        PrintExpression(domain->right);
        break;
    case LESSER_LESSER_DOMAIN:
    case LESSER_LESSER_EQUAL_DOMAIN:
    case LESSER_EQUAL_LESSER_EQUAL_DOMAIN:
    case LESSER_EQUAL_LESSER_DOMAIN:
        LogError("Invalid boolean comparison");
    default:
        abort();
        break;
    }
}

void PrintDomainForFunction(Domain * domain) {
    switch (domain->type) {
    case GREATER_DOMAIN:
        PrintExpression(domain->left);
        Print(" > ");
        PrintExpression(domain->right);
        break;
    case GREATER_EQUAL_DOMAIN:
        PrintExpression(domain->left);
        Print(" >= ");
        PrintExpression(domain->right);
        break;
    case LESSER_EQUAL_DOMAIN:
        PrintExpression(domain->left);
        Print(" <= ");
        PrintExpression(domain->right);
        break;
    case LESSER_DOMAIN:
        PrintExpression(domain->left);
        Print(" < ");
        PrintExpression(domain->right);
        break;
    case EQUAL_DOMAIN:
        PrintExpression(domain->left);
        Print(" == ");
        PrintExpression(domain->right);
        break;
    case NOT_EQUAL_DOMAIN:
        PrintExpression(domain->left);
        Print(" != ");
        PrintExpression(domain->right);
        break;
    case LESSER_LESSER_DOMAIN:
        PrintExpression(domain->left);
        Print(" < ");
        PrintExpression(domain->mid);
        Print(" && ");
        PrintExpression(domain->mid);
        Print(" < ");
        PrintExpression(domain->right);
        break;
    case LESSER_LESSER_EQUAL_DOMAIN:
        PrintExpression(domain->left);
        Print(" < ");
        PrintExpression(domain->mid);
        Print(" && ");
        PrintExpression(domain->mid);
        Print(" <= ");
        PrintExpression(domain->right);
        break;
    case LESSER_EQUAL_LESSER_EQUAL_DOMAIN:
        PrintExpression(domain->left);
        Print(" <= ");
        PrintExpression(domain->mid);
        Print(" && ");
        PrintExpression(domain->mid);
        Print(" <= ");
        PrintExpression(domain->right);
        break;
    case LESSER_EQUAL_LESSER_DOMAIN:
        PrintExpression(domain->left);
        Print(" <= ");
        PrintExpression(domain->mid);
        Print(" && ");
        PrintExpression(domain->mid);
        Print(" < ");
        PrintExpression(domain->right);
        break;
    }
}

void PrintBoolean(Boolean * boolean) {
    switch (boolean->type) {
    case DOMAIN_BOOLEAN:
        PrintDomain(boolean->domain);
        break;
    case OR_BOOLEAN:
        PrintBoolean(boolean->left);
        Print(" || ");
        PrintBoolean(boolean->right);
        break;
    case AND_BOOLEAN:
        PrintBoolean(boolean->left);
        Print(" && ");
        PrintBoolean(boolean->right);
        break;
    case NOT_BOOLEAN:
        Print("!");
        PrintBoolean(boolean->left);
        break;
    case BOOLEAN:
        Print("(");
        PrintBoolean(boolean->left);
        Print(")");
        break;
    default:
        abort();
    }
}

void PrintIf(If * if_st) {
    Print(" if (");
    PrintBoolean(if_st->boolean);
    Print(") ");
    PrintBlock(if_st->block);
    Print(";");
    if (if_st->else_block != NULL) {
        Print(" else ");
        PrintBlock(if_st->else_block);
    }
    Print("endif");
}

void PrintFor(For * for_st) {
    if (for_st->init_variable != NULL) {
        PrintDeclareVariable(for_st->init_variable);
        Print(";");
    }
    Print(" while (true)");
    Print("if (!(");
    PrintBoolean(for_st->boolean);
    Print(")) break; endif; ");
    PrintAssignVariable(for_st->action_variable);
    Print(";");
    PrintBlock(for_st->block);
    Print(" endwhile");
}


void PrintBlock(Block * block) {
    Statement * statement = block->statement;
    while (statement != NULL) {
        PrintStatement(statement);
        statement = statement->next;
    }
}

void PrintDeclareVariable(Variable * variable) {
    Print("%s", variable->name);
    if (variable->value != NULL) {
        Print(" = ");
        PrintExpression(variable->value);
    }
}

void PrintAssignVariable(AssignVariable * variable) {
    switch (variable->type) {
        case ADD_ASSIGN_EXPRESSION:
            Print("%s += ", variable->id);
            PrintExpression(variable->expression);
            break;
        case SUB_ASSIGN_EXPRESSION:
            Print("%s -= ", variable->id);
            PrintExpression(variable->expression);
            break;
        case MUL_ASSIGN_EXPRESSION:
            Print("%s *= ", variable->id);
            PrintExpression(variable->expression);
            break;
        case DIV_ASSIGN_EXPRESSION:
            Print("%s /= ", variable->id);
            PrintExpression(variable->expression);
            break;
        case INCREMENT_EXPRESSION:
            Print("%s += 1", variable->id);
            break;
        case DECREMENT_EXPRESSION:
            Print("%s -= 1", variable->id);
            break;
        case ASSIGN_EXPRESSION:
            Print("%s = ", variable->id);
            PrintExpression(variable->expression);
            break;
        default:
            abort();
    }
}

void PrintPrint(Print * print_st) {
    Print("printf(");
    switch (print_st->type) {
        case STRING_PRINT:
            Print("%s", print_st->string);
            break;
        case EXPRESSION_PRINT:
            Print("\"%%f\\n\",");
            PrintExpression(print_st->expression);
            break;
        default:
            abort();
    }
    Print(")");
}

void PrintWhile(While * while_st) {
    Print("while (");
    PrintBoolean(while_st->boolean);
    Print(") ");
    PrintBlock(while_st->block);
    Print(" endwhile");
}

void PrintCompositeFunction(Composite * composite, char * flag) {
    Print("%s(", composite->outter_func);
    char flagAux = 0;
    switch (composite->type) {
        case ID_COMPOSITE:
            Print("%s(x))", composite->inner_func);
            break;
        case COMPOSITE_COMPOSITE:
            PrintCompositeFunction(composite->inner_comp, &flagAux);
            if (*flag)
                Print(";");
            break;
        default:
            LogError("Invalid composite function");
            abort();
            break;
    }
}

void PrintDomainedFunction(Function * function) {
    Print("(");
    PrintDomainForFunction(function->domain);
    Print(") * ");
    PrintExpression(function->expression);
}

void PrintDeclareFunction(Function * function) {
    Function * currFunc = function;
    Print("%s=function_handle( ", function->name);
    char flag = 1;
    while (currFunc != NULL) {
        switch (currFunc->type) {
            case EXPRESSION_FUNCTION:
                PrintExpression(currFunc->expression);
                break;
            case EXPRESSION_DOMAIN_FUNCTION:
                PrintDomainedFunction(currFunc);
                break;
            case COMPOSITE_FUNCTION:
                PrintCompositeFunction(currFunc->composite, &flag);
                break;
            default:
                LogError("Incorrect function definition");
                abort();
        }
        currFunc = currFunc->next;
        if (currFunc != NULL)
            Print(" + ");
    }
    Print(")");
}

void PrintStatement(Statement * statement) {
    switch (statement->type) {
        case WHILE_STATEMENT:
            PrintWhile(statement->while_st);
            break;
        case FOR_STATEMENT:
            PrintFor(statement->for_st);
            break;
        case IF_STATEMENT:
            PrintIf(statement->if_st);
            break;
        case PRINT_STATEMENT:
            PrintPrint(statement->print);
            break;
        case INTEGRATE_STATEMENT:
            PrintIntegrate(statement->integrate);
            break;
        case DERIVATIVE_STATEMENT:
            PrintDerivative(statement->derivative);
            break;
        case EVALUATE_STATEMENT:
            PrintEvaluate(statement->evaluate);
            break;
        case DECLARE_VARIABLE_STATEMENT:
            PrintDeclareVariable(statement->declareVariable);
            break;
        case ASSIGN_VARIABLE_STATEMENT:
            PrintAssignVariable(statement->assignVariable);
            break;
        case DECLARE_FUNCTION_STATEMENT:
            PrintDeclareFunction(statement->declareFunction);
            break;
        default:
            LogError("Invalid statement");
            abort();
    }
    Print(";\n");
}

void Generator(Program * program, FILE * fileAux) {
    file = fileAux;
    
    Statement * currStatement = program->statement;
    Print("pkg load symbolic;\n");
    Print("sympref quiet on; syms x;\n");
    while (currStatement != NULL) {
        PrintStatement(currStatement);
        currStatement = currStatement->next;
    }
}
