#include "../support/nodes.h"
#include "../support/logger.h"
#include <stdio.h>
#include <string.h>
#include <stdbool.h>

#define DEFAULT_SIZE 20

typedef struct Table {
    char ** elements;
    int used;
    int size;
    struct Table * prev;
} Table;

typedef struct Stack {
    Table * top;
} Stack;

bool AnalyzeDerivative(Derivative * derivative, Stack * stack);
bool AnalyzeStatement(Statement * statement, Stack * stack);
bool AnalyzeEvaluate(Evaluate * evaluate, Stack * stack);
bool AnalyzeIntegrate(Integrate * integrate, Stack * stack);
bool AnalyzeFunctionDerivative(Derivative * derivative, Stack * stack);
bool AnalyzeFunctionStatement(Statement * statement, Stack * stack);
bool AnalyzeFunctionEvaluate(Evaluate * evaluate, Stack * stack);
bool AnalyzeFunctionIntegrate(Integrate * integrate, Stack * stack);

bool FindVar(Table * table, char * name) {
    for (int i = 0; i < table->used; i++) {
        if (!strcmp(name, table->elements[i]))
            return true;
    }
    return false;
}

void CopyTable(Table * src, Table * dest) {
    dest->size = src->size;
    dest->used = src->used;
    
    for (int i = 0; i < src->used; i++) {
        dest->elements[i] = malloc(strlen(src->elements[i]) + 1);
        strcpy(dest->elements[i], src->elements[i]);
    }
}

Table * PushNewTable(Stack * stack) {
    Table * newTable;
    newTable = malloc(sizeof(Table));
    if (stack->top != NULL) {
        newTable->elements = malloc(stack->top->size * sizeof(char *));
        CopyTable(stack->top, newTable);
    }
    else {
        newTable->elements = malloc(DEFAULT_SIZE * sizeof(char *));
        newTable->size = DEFAULT_SIZE;
        newTable->used = 0;
    }
    newTable->prev = stack->top;
    stack->top = newTable;
    return newTable;
}

void PopTable(Stack * stack) {
    if (stack->top == NULL)
        return;
    Table * poppedTable = stack->top;
    stack->top = poppedTable->prev;
    for (int i = 0; i < poppedTable->used; i++) {
        free(poppedTable->elements[i]);
    }
    free(poppedTable->elements);
    free(poppedTable);
}

void AddVar(Table * table, char * varName) {
    if (table->used == table->size) {
        table->elements = realloc(table->elements, table->size * 2);
        table->size *= 2;
    }
    table->elements[table->used] = malloc(strlen(varName) + 1);
    strcpy(table->elements[table->used++], varName);
}

bool AnalyzeComposite(Composite * composite, Stack * stack) {
    Table * table = stack->top;
    if (table == NULL) {
        LogError("Semantic analyzer error");
        return false;
    }
    if (!FindVar(table, composite->outter_func)) {
        LogError("Function %s not defined", composite->outter_func);
        return false;
    }
    if (composite->type == COMPOSITE_COMPOSITE)
        return AnalyzeComposite(composite->inner_comp, stack);
    else {
        if (!FindVar(table, composite->inner_func)) {
            LogError("Variable %s not defined", composite->inner_func);
            return false;
        }
    }
    return true;
}

bool AnalyzeExpression(Expression * expression, Stack * stack) {
    Table * table = stack->top;
    if (table == NULL) {
        LogError("Semantic analyzer error");
        return false;
    }
    switch (expression->type) {
        case DERIVATIVE_EXPRESSION:
            return AnalyzeDerivative(expression->derivative, stack);       
        case INTEGRATE_EXPRESSION:
            return AnalyzeIntegrate(expression->integrate, stack);
        case EVALUATE_EXPRESSION:
            return AnalyzeEvaluate(expression->evaluate, stack);
        case NEGATIVE_EXPRESSION:
        case PARENTHESIS_EXPRESSION:
            return AnalyzeExpression(expression->leftExpression, stack);
        case NUMBER_EXPRESSION:
            break;
        case ADD_EXPRESSION:
        case SUB_EXPRESSION:
        case MUL_EXPRESSION:
        case DIV_EXPRESSION:
        case POW_EXPRESSION:
            if (!AnalyzeExpression(expression->leftExpression, stack))
                return false;
            return AnalyzeExpression(expression->rightExpression, stack);
        case ID_EXPRESSION:
            if (!FindVar(table, expression->id)) {
                LogError("Variable %s not defined", expression->id);
                return false;
            }
            break;
        default:
            return false;
    }
    return true;
}

bool AnalyzeFunctionExpression(Expression * expression, Stack * stack) {
    Table * table = stack->top;
    if (table == NULL) {
        LogError("Semantic analyzer error");
        return false;
    }
    switch (expression->type) {
        case DERIVATIVE_EXPRESSION:
            return AnalyzeFunctionDerivative(expression->derivative, stack);
        case INTEGRATE_EXPRESSION:
            return AnalyzeFunctionIntegrate(expression->integrate, stack);
        case EVALUATE_EXPRESSION:
            return AnalyzeFunctionEvaluate(expression->evaluate, stack);
        case NEGATIVE_EXPRESSION:
        case PARENTHESIS_EXPRESSION:
            return AnalyzeFunctionExpression(expression->leftExpression, stack);
        case NUMBER_EXPRESSION:
            break;
        case ADD_EXPRESSION:
        case SUB_EXPRESSION:
        case MUL_EXPRESSION:
        case DIV_EXPRESSION:
        case POW_EXPRESSION:
            if (!AnalyzeFunctionExpression(expression->leftExpression, stack))
                return false;
            return AnalyzeFunctionExpression(expression->rightExpression, stack);
        case ID_EXPRESSION:
            if (strcmp(expression->id, "x")) {
                LogError("Function expression must only have \'x\' as a variable");
                return false;
            }
            break;
        default:
            LogError("Invalid expression");
            return false;
    }
    return true;
}

bool AnalyzeFunctionDomain(Domain * domain, Stack * stack) {
    switch (domain->type) {
        case GREATER_DOMAIN:
        case GREATER_EQUAL_DOMAIN:
        case LESSER_EQUAL_DOMAIN:
        case LESSER_DOMAIN:
        case EQUAL_DOMAIN:
        case NOT_EQUAL_DOMAIN:
            if (!AnalyzeFunctionExpression(domain->left, stack))
                return false;
            return AnalyzeFunctionExpression(domain->right, stack);
        case LESSER_LESSER_DOMAIN:
        case LESSER_LESSER_EQUAL_DOMAIN:
        case LESSER_EQUAL_LESSER_EQUAL_DOMAIN:
        case LESSER_EQUAL_LESSER_DOMAIN:
            if (!AnalyzeFunctionExpression(domain->left, stack))
                return false;
            if (!AnalyzeFunctionExpression(domain->mid, stack))
                return false;
            return AnalyzeFunctionExpression(domain->right, stack);
        default:
            LogError("Invalid domain");
            return false;
    }
    return true;
}

bool AnalyzeDomain(Domain * domain, Stack * stack) {
    switch (domain->type) {
        case GREATER_DOMAIN:
        case GREATER_EQUAL_DOMAIN:
        case LESSER_EQUAL_DOMAIN:
        case LESSER_DOMAIN:
        case EQUAL_DOMAIN:
        case NOT_EQUAL_DOMAIN:
            if (!AnalyzeExpression(domain->left, stack))
                return false;
            return AnalyzeExpression(domain->right, stack);
        case LESSER_LESSER_DOMAIN:
        case LESSER_LESSER_EQUAL_DOMAIN:
        case LESSER_EQUAL_LESSER_EQUAL_DOMAIN:
        case LESSER_EQUAL_LESSER_DOMAIN:
            break;
        default:
            LogError("Invalid domain");
            return false;
    }
    return true;
}

bool AnalyzeBoolean(Boolean * boolean, Stack * stack) {
    switch (boolean->type) {
        case DOMAIN_BOOLEAN:
        case NOT_BOOLEAN:
        case BOOLEAN:
            if (!AnalyzeDomain(boolean->domain, stack))
                return false;
            break;
        case OR_BOOLEAN:
        case AND_BOOLEAN:
            if (!AnalyzeBoolean(boolean->left, stack))
                return false;
            return AnalyzeBoolean(boolean->right, stack);
        default:
            LogError("Invalid boolean");
            return false;
    }
    return true;
}

bool AnalyzeDeclareVariable(Variable * variable, Stack * stack) {
    Table * table = stack->top;
    if (table == NULL) {
        LogError("Semantic analyzer error");
        return false;
    }
    if (!strcmp(variable->name, "x")) {
        LogError("Cannot use variable name 'x' outside of function");
        return false;
    }
    if (!FindVar(table, variable->name)) {
        AddVar(table, variable->name);
    }
    return true;
}

bool AnalyzeDeclareFunction(Function * function, Stack * stack) {
    Table * table = stack->top;
    if (table == NULL) {
        LogError("Semantic analyzer error");
        return false;
    }
    if (function->name != NULL) {
        if (!strcmp(function->name, "x")) {
            LogError("Cannot use function name 'x'");
            return false;
        }
        if (!FindVar(table, function->name)) {
            AddVar(table, function->name);
        }
    }
    switch (function->type) {
        case EXPRESSION_FUNCTION:
            if (!AnalyzeFunctionExpression(function->expression, stack))
                return false;
            break;
        case EXPRESSION_DOMAIN_FUNCTION:
            if (!AnalyzeFunctionDomain(function->domain, stack))
                return false;
            if (!AnalyzeFunctionExpression(function->expression, stack))
                return false;
            break;
        case COMPOSITE_FUNCTION:
            if (!AnalyzeComposite(function->composite, stack))
                return false;
            break;
        default:
            LogError("Invalid function");
            return false;
    }
    if (function->next != NULL)
        return AnalyzeDeclareFunction(function->next, stack);
    return true;
}

bool AnalyzeAssignVariable(AssignVariable * assign, Stack * stack) {
    Table * table = stack->top;
    if (table == NULL) {
        LogError("Semantic analyzer error");
        return false;
    }
    if (!FindVar(table, assign->id)) {
        LogError("Variable %s not defined", assign->id);
        return false;
    }

    return true;
}

bool AnalyzeBlock(Block * block, Stack * stack) {
    PushNewTable(stack);

    Statement * currStatement = block->statement;
    while (currStatement != NULL) {
        if (!AnalyzeStatement(currStatement, stack))
            return false;
        currStatement = currStatement->next;
    }

    PopTable(stack);
    return true;
}

bool AnalyzeWhile(While * while_st, Stack * stack) {
    return AnalyzeBlock(while_st->block, stack);
}

bool AnalyzeIf(If * if_st, Stack * stack) {
    AnalyzeBlock(if_st->block, stack);
    if (if_st->else_block != NULL)
        return AnalyzeBlock(if_st->else_block, stack);
    return true;
}

bool AnalyzeFor(For * for_st, Stack * stack) {
    if (for_st->init_variable != NULL) {
        if (!AnalyzeDeclareVariable(for_st->init_variable, stack))
            return false;
    }
    if (!AnalyzeAssignVariable(for_st->action_variable, stack))
        return false;
    return AnalyzeBlock(for_st->block, stack);
}

bool AnalyzePrint(Print * print, Stack * stack) {
    if (print->type == EXPRESSION_PRINT)
        return AnalyzeExpression(print->expression, stack);
    return true;
}

bool AnalyzeIntegrate(Integrate * integrate, Stack * stack) {
    Table * table = stack->top;
    if (table == NULL) {
        LogError("Semantic analyzer error");
        return false;
    }
    if (!FindVar(table, integrate->function)) {
        LogError("Function %s not defined", integrate->function);
        return false;
    }
    if (!AnalyzeExpression(integrate->lvalue, stack))
        return false;
    return AnalyzeExpression(integrate->rvalue, stack);
}

bool AnalyzeFunctionIntegrate(Integrate * integrate, Stack * stack) {
    Table * table = stack->top;
    if (table == NULL) {
        LogError("Semantic analyzer error");
        return false;
    }
    if (!FindVar(table, integrate->function)) {
        LogError("Function %s not defined", integrate->function);
        return false;
    }
    if (!AnalyzeFunctionExpression(integrate->lvalue, stack))
        return false;
    return AnalyzeFunctionExpression(integrate->rvalue, stack);
}

bool AnalyzeDerivative(Derivative * derivative, Stack * stack) {
    Table * table = stack->top;
    if (table == NULL) {
        LogError("Semantic analyzer error");
        return false;
    }
    if (!FindVar(table, derivative->function)) {
        LogError("Function %s not defined", derivative->function);
        return false;
    }
    return AnalyzeExpression(derivative->value, stack);
}

bool AnalyzeFunctionDerivative(Derivative * derivative, Stack * stack) {
    Table * table = stack->top;
    if (table == NULL) {
        LogError("Semantic analyzer error");
        return false;
    }
    if (!FindVar(table, derivative->function)) {
        LogError("Function %s not defined", derivative->function);
        return false;
    }
    return AnalyzeFunctionExpression(derivative->value, stack);
}

bool AnalyzeEvaluate(Evaluate * evaluate, Stack * stack) {
    Table * table = stack->top;
    if (table == NULL) {
        LogError("Semantic analyzer error");
        return false;
    }
    if (!FindVar(table, evaluate->function)) {
        LogError("Function %s not defined", evaluate->function);
        return false;
    }
    return AnalyzeExpression(evaluate->value, stack);
}

bool AnalyzeFunctionEvaluate(Evaluate * evaluate, Stack * stack) {
    Table * table = stack->top;
    if (table == NULL) {
        LogError("Semantic analyzer error");
        return false;
    }
    if (!FindVar(table, evaluate->function)) {
        LogError("Function %s not defined", evaluate->function);
        return false;
    }
    return AnalyzeFunctionExpression(evaluate->value, stack);
}

bool AnalyzeStatement(Statement * statement, Stack * stack) {
    Table * table = stack->top;
    if (table == NULL) {
        LogError("Semantic analyzer error");
        return false;

    }
    switch (statement->type) {
        case WHILE_STATEMENT:
            return AnalyzeWhile(statement->while_st, stack);
        case FOR_STATEMENT:
            return AnalyzeFor(statement->for_st, stack);
        case IF_STATEMENT:
            return AnalyzeIf(statement->if_st, stack);
        case PRINT_STATEMENT:
            return AnalyzePrint(statement->print, stack);
        case INTEGRATE_STATEMENT:
            return AnalyzeIntegrate(statement->integrate, stack);
        case DERIVATIVE_STATEMENT:
            return AnalyzeDerivative(statement->derivative, stack);
        case EVALUATE_STATEMENT:
            return AnalyzeEvaluate(statement->evaluate, stack);
        case DECLARE_VARIABLE_STATEMENT:
            return AnalyzeDeclareVariable(statement->declareVariable, stack);
        case ASSIGN_VARIABLE_STATEMENT:
            return AnalyzeAssignVariable(statement->assignVariable, stack);
        case DECLARE_FUNCTION_STATEMENT:
            return AnalyzeDeclareFunction(statement->declareFunction, stack);
        default:
            LogError("Invalid statement");
            return false;
    }
    return true;
}

bool Analyze(Program * program) {
    Statement * currStatement = program->statement;
    Stack * stack = calloc(1, sizeof(Stack));
    bool success = true;
    PushNewTable(stack);
    while (currStatement != NULL) {
        if (!AnalyzeStatement(currStatement, stack)) {
            success = false;
            break;
        }
        currStatement = currStatement->next;
    }
    while (stack->top != NULL)
        PopTable(stack);
    free(stack);
    return success;
}
